//
// Huffyuv v1.2.3, by Ben Rudiak-Gould.
// http://www.math.berkeley.edu/~benrg/huffyuv.html
//
// Based on MSYUV sample code, which is:
// Copyright (c) 1993 Microsoft Corporation.
// All Rights Reserved.
//
// Changes copyright 2000 Ben Rudiak-Gould, and distributed under
// the terms of the GNU General Public License, v2 or later.  See
// http://www.gnu.org/copyleft/gpl.html.
//
// I edit these files in 10-point Verdana, a proportionally-spaced font.
// You may notice formatting oddities if you use a monospaced font.
//


#include "huffyuv.h"
#include "resource.h"


TCHAR szDescription[] = TEXT("Huffyuv v1.2.3");
TCHAR szName[]        = TEXT("Huffyuv");

#define VERSION         0x00010002      // 1.2


/********************************************************************
********************************************************************/

void Msg(const char fmt[], ...) {
  static int debug = GetPrivateProfileInt("debug", "log", 0, "huffyuv.ini");
  if (!debug) return;

  DWORD written;
  char buf[2000];
  va_list val;
  
  va_start(val, fmt);
  wvsprintf(buf, fmt, val);

  const COORD _80x50 = {80,50};
  static BOOL startup = (AllocConsole(), SetConsoleScreenBufferSize(GetStdHandle(STD_OUTPUT_HANDLE), _80x50));
  WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), buf, lstrlen(buf), &written, 0);
}


int AppFlags() {
  static int flags = -1;
  if (flags < 0) {
    flags = 0;
    TCHAR apppath[MAX_PATH];
    if (GetModuleFileName(NULL, apppath, MAX_PATH)) {
      TCHAR* appname = strrchr(apppath, '\\');
      appname = appname ? appname+1 : apppath;
      Msg("App name is %s; ", appname);
      if (!lstrcmpi(appname, TEXT("premiere.exe")))
        flags = 1;
      if (!lstrcmpi(appname, TEXT("veditor.exe")))
        flags = 1;
      if (!lstrcmpi(appname, TEXT("afterfx.exe")))
        flags = 2;
      Msg("flags=%d\n", flags);
    }
  }
  return flags;
}


/********************************************************************
********************************************************************/

CodecInst* Open(ICOPEN* icinfo) {
  if (icinfo && icinfo->fccType != ICTYPE_VIDEO)
      return NULL;

  CodecInst* pinst = new CodecInst();

  if (icinfo) icinfo->dwError = pinst ? ICERR_OK : ICERR_MEMORY;

  return pinst;
}

DWORD Close(CodecInst* pinst) {
//    delete pinst;       // this caused problems when deleting at app close time
    return 1;
}

/********************************************************************
********************************************************************/

BOOL CodecInst::QueryAbout() { return TRUE; }

static INT_PTR CALLBACK AboutDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  if (uMsg == WM_COMMAND) {
    switch (LOWORD(wParam)) {
    case IDOK:
      EndDialog(hwndDlg, 0);
      break;
    case IDC_HOMEPAGE:
      ShellExecute(NULL, NULL, "http://www.math.berkeley.edu/~benrg/huffyuv.html", NULL, NULL, SW_SHOW);
      break;
    case IDC_EMAIL:
      ShellExecute(NULL, NULL, "mailto:benrg@math.berkeley.edu", NULL, NULL, SW_SHOW);
      break;
    }
  }
  return FALSE;
}

DWORD CodecInst::About(HWND hwnd) {
  DialogBox(hmoduleHuffyuv, MAKEINTRESOURCE(IDD_ABOUT), hwnd, AboutDialogProc);
  return ICERR_OK;
}

static INT_PTR CALLBACK ConfigureDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  if (uMsg == WM_INITDIALOG) {
    CheckDlgButton(hwndDlg, IDC_COMPRESSASYUY2,
      GetPrivateProfileInt("general", "compress_as_yuy2", 0, "huffyuv.ini") ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton(hwndDlg, IDC_RGBOUTPUT,
      GetPrivateProfileInt("debug", "rgboutput", 0, "huffyuv.ini") ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton(hwndDlg, IDC_LOG,
      GetPrivateProfileInt("debug", "log", 0, "huffyuv.ini") ? BST_CHECKED : BST_UNCHECKED);
  } else if (uMsg == WM_COMMAND) {
    switch (LOWORD(wParam)) {
    case IDOK:
      WritePrivateProfileString("general", "compress_as_yuy2",
        (IsDlgButtonChecked(hwndDlg, IDC_COMPRESSASYUY2) == BST_CHECKED) ? "1" : "0", "huffyuv.ini");
      WritePrivateProfileString("debug", "rgboutput",
        (IsDlgButtonChecked(hwndDlg, IDC_RGBOUTPUT) == BST_CHECKED) ? "1" : NULL, "huffyuv.ini");
      WritePrivateProfileString("debug", "log",
        (IsDlgButtonChecked(hwndDlg, IDC_LOG) == BST_CHECKED) ? "1" : "0", "huffyuv.ini");
    case IDCANCEL:
      EndDialog(hwndDlg, 0);
      break;
    default:
      return AboutDialogProc(hwndDlg, uMsg, wParam, lParam);    // handle email and home-page buttons
    }
  }
  return FALSE;
}

BOOL CodecInst::QueryConfigure() { return TRUE; }

DWORD CodecInst::Configure(HWND hwnd) {
  DialogBox(hmoduleHuffyuv, MAKEINTRESOURCE(IDD_CONFIGURE), hwnd, ConfigureDialogProc);
  return ICERR_OK;
}


// we have no state information which needs to be stored

DWORD CodecInst::GetState(LPVOID pv, DWORD dwSize) { return 0; }

DWORD CodecInst::SetState(LPVOID pv, DWORD dwSize) { return 0; }


DWORD CodecInst::GetInfo(ICINFO* icinfo, DWORD dwSize) {
  if (icinfo == NULL)
    return sizeof(ICINFO);

  if (dwSize < sizeof(ICINFO))
    return 0;

  icinfo->dwSize            = sizeof(ICINFO);
  icinfo->fccType           = ICTYPE_VIDEO;
  icinfo->fccHandler        = FOURCC_HFYU;
  icinfo->dwFlags           = 0;

  icinfo->dwVersion         = VERSION;
  icinfo->dwVersionICM      = ICVERSION;
  MultiByteToWideChar(CP_ACP, 0, szDescription, -1, icinfo->szDescription, sizeof(icinfo->szDescription)/sizeof(WCHAR));
  MultiByteToWideChar(CP_ACP, 0, szName, -1, icinfo->szName, sizeof(icinfo->szName)/sizeof(WCHAR));

  return sizeof(ICINFO);
}


/********************************************************************
********************************************************************/


struct PrintBitmapType {
  char s[32];
  PrintBitmapType(LPBITMAPINFOHEADER lpbi) {
    if (!lpbi)
      strcpy(s,  "(null)");
    else {
      *(DWORD*)s = lpbi->biCompression;
      s[4] = 0;
      if (!isalnum(s[0]) || !isalnum(s[1]) || !isalnum(s[2]) || !isalnum(s[3]))
        wsprintfA(s, "%x", lpbi->biCompression);
      wsprintfA(strchr(s, 0), ", bits = %d", lpbi->biBitCount);
    }
  }
};


// fast clipping of values to unsigned char range

static unsigned char clip[896];

static void InitClip() {
  memset(clip, 0, 320);
  for (int i=0; i<256; ++i) clip[i+320] = i;
  memset(clip+320+256, 255, 320);
}

static inline unsigned char Clip(int x)
  { return clip[320 + ((x+0x8000) >> 16)]; }


/********************************************************************
********************************************************************/


// 0=unknown, -1=compressed YUY2, -2=compressed RGB, -3=compressed RGBA, 1=YUY2, 2=UYVY, 3=RGB 24-bit, 4=RGB 32-bit
static int GetBitmapType(LPBITMAPINFOHEADER lpbi) {
  if (!lpbi)
    return 0;
  if (lpbi->biCompression == FOURCC_VYUY || lpbi->biCompression == FOURCC_YUY2)
    return 1;
  if (lpbi->biCompression == FOURCC_UYVY)
    return 2;
  if (lpbi->biCompression == 0 || lpbi->biCompression == ' BID')
    return (lpbi->biBitCount == 24) ? 3 : (lpbi->biBitCount == 32) ? 4 : 0;
  if (lpbi->biCompression == FOURCC_HFYU)
    return (lpbi->biBitCount == 16) ? -1 : (lpbi->biBitCount == 24) ? -2 : (lpbi->biBitCount == 32) ? -3 : 0;
  return 0;
}


static bool CanCompress(LPBITMAPINFOHEADER lpbiIn) {
  int intype = GetBitmapType(lpbiIn);
  return (intype == 1 || intype == 2 || intype == 3 || (intype == 4 && (AppFlags()&2)));
}


static bool CanCompress(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  if (!lpbiOut) return CanCompress(lpbiIn);
  int outtype = GetBitmapType(lpbiOut);
  switch (GetBitmapType(lpbiIn)) {
    case 1: case 2:
      return (outtype == -1);
    case 3:
      return (outtype == 1 || outtype == -1 || outtype == -2);
    case 4:
      return (outtype == -3 && (AppFlags() & 2));
    default:
      return false;
  }
}


/********************************************************************
********************************************************************/


DWORD CodecInst::CompressQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  Msg("CompressQuery: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));
  return CanCompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}


DWORD CodecInst::CompressGetFormat(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  if (lpbiOut == 0)
    return sizeof(BITMAPINFOHEADER);

  if (!CanCompress(lpbiIn))
    return ICERR_BADFORMAT;

  bool compress_as_yuy2 = !!GetPrivateProfileInt("general", "compress_as_yuy2", false, "huffyuv.ini");

  *lpbiOut = *lpbiIn;
  lpbiOut->biCompression = FOURCC_HFYU;
  int intype = GetBitmapType(lpbiIn);
  lpbiOut->biBitCount = (intype==4) ? 32 : (intype==1 || intype==2 || compress_as_yuy2) ? 16 : 24;
  return ICERR_OK;
}


DWORD CodecInst::CompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
//  total_bpp = 0;
//  num_frames = 0;
  Msg("CompressBegin: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));

  CompressEnd();  // free resources if necessary

  // allocate buffer if compressing RGB->YUY2->HFYU
  if (GetBitmapType(lpbiIn) == 3 && GetBitmapType(lpbiOut) == -1)
    yuy2_buffer = new unsigned char[lpbiIn->biWidth * lpbiIn->biHeight * 2];

  InitClip();

  return CanCompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}

DWORD CodecInst::CompressGetSize(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  // Assume 24 bpp worst-case for YUY2 input, 40 bpp for RGB.
  // The actual worst case is 43/51 bpp currently, but this is exceedingly improbable
  // (probably impossible with real captured video)
  return lpbiIn->biWidth * lpbiIn->biHeight * ((GetBitmapType(lpbiIn) <= 2) ? 3 : 5);
}


DWORD CodecInst::Compress(ICCOMPRESS* icinfo, DWORD dwSize) {
  int intype = GetBitmapType(icinfo->lpbiInput);
  int outtype = GetBitmapType(icinfo->lpbiOutput);

  if ((intype == 1 || intype == 3) && outtype == -1)
    return CompressYUY2(icinfo, dwSize);
  if (intype == 2 && outtype == -1)
    return CompressUYVY(icinfo, dwSize);
  else if (intype == 3 && outtype == -2)
    return CompressRGB(icinfo, dwSize);
  else if (intype == 3 && outtype == 1) {
    ConvertRGB24toYUY2((unsigned char*)icinfo->lpInput, (unsigned char*)icinfo->lpOutput,
      icinfo->lpbiInput->biWidth, icinfo->lpbiInput->biHeight);
    return ICERR_OK;
  } else if (intype == 4 && outtype == -3)
    return CompressRGBA(icinfo, dwSize);
  else
    return ICERR_BADFORMAT;
}


#define HUFF(table,index,spacing) \
do { \
  const huffman_entry& huff = table[((int)in[index] - in[index-spacing]) & 255]; \
  bit_count_minus_32 += huff.shift; \
  if (bit_count_minus_32 < 0) { \
    bits = (bits << huff.shift) | huff.add; \
  } else { \
    *out++ = (bits << (huff.shift-bit_count_minus_32)) | (huff.add >> bit_count_minus_32); \
    bits = huff.add; \
    bit_count_minus_32 -= 32; \
  } \
} while(0)


DWORD CodecInst::CompressYUY2(ICCOMPRESS* icinfo, DWORD dwSize) {
  // if compressing from RGB, convert to YUY2 first, using the temporary
  // buffer which had better have been allocated for us.
  unsigned char* in = (unsigned char*)icinfo->lpInput;
  unsigned char* end = in + icinfo->lpbiInput->biSizeImage;
  if (GetBitmapType(icinfo->lpbiInput) == 3) {
    ConvertRGB24toYUY2(in, yuy2_buffer, icinfo->lpbiInput->biWidth, icinfo->lpbiInput->biHeight);
    in = yuy2_buffer;
    end = in + icinfo->lpbiInput->biWidth * icinfo->lpbiInput->biHeight * 2;
  }

  DWORD* out = (DWORD*)icinfo->lpOutput;
  *out++ = *(DWORD*)in;

  DWORD bits = 0;
  int bit_count_minus_32 = -32;
  in += 4;
  do {
    HUFF(luma,0,2);   // y
    HUFF(chroma,1,4); // u/v
    in += 2;
  } while (in < end);

  if (-bit_count_minus_32 < 32)
    *out++ = bits << (-bit_count_minus_32);

  icinfo->lpbiOutput->biSizeImage = DWORD(out) - DWORD(icinfo->lpOutput);
  if (icinfo->lpckid)
    *icinfo->lpckid = FOURCC_HFYU;
  *icinfo->lpdwFlags = AVIIF_KEYFRAME;

  return ICERR_OK;
}


// this is almost a copy of CompressYUY2
DWORD CodecInst::CompressUYVY(ICCOMPRESS* icinfo, DWORD dwSize) {
  unsigned char* in = (unsigned char*)icinfo->lpInput;
  unsigned char* end = in + icinfo->lpbiInput->biSizeImage;

  DWORD* out = (DWORD*)icinfo->lpOutput;
  *out++ = ((in[2]*256+in[3])*256+in[0])*256+in[1];

  DWORD bits = 0;
  int bit_count_minus_32 = -32;
  in += 4;
  do {
    HUFF(luma,1,2);   // y
    HUFF(chroma,0,4); // u/v
    in += 2;
  } while (in < end);

  if (-bit_count_minus_32 < 32)
    *out++ = bits << (-bit_count_minus_32);

  icinfo->lpbiOutput->biSizeImage = DWORD(out) - DWORD(icinfo->lpOutput);
  if (icinfo->lpckid)
    *icinfo->lpckid = FOURCC_HFYU;
  *icinfo->lpdwFlags = AVIIF_KEYFRAME;

  return ICERR_OK;
}


DWORD CodecInst::CompressRGB(ICCOMPRESS* icinfo, DWORD dwSize) {
  unsigned char* in = (unsigned char*)icinfo->lpInput;
  DWORD* out = (DWORD*)icinfo->lpOutput;
  *out++ = *(DWORD*)in << 8;

  DWORD bits = 0;
  int bit_count_minus_32 = -32;
  unsigned char* end = in + icinfo->lpbiInput->biSizeImage;
  in += 3;
  do {
    HUFF(luma,0,3);
    ++in;
  } while (in < end);

  if (-bit_count_minus_32 < 32)
    *out++ = bits << (-bit_count_minus_32);

  icinfo->lpbiOutput->biSizeImage = DWORD(out) - DWORD(icinfo->lpOutput);
  if (icinfo->lpckid)
    *icinfo->lpckid = FOURCC_HFYU;
  *icinfo->lpdwFlags = AVIIF_KEYFRAME;

  return ICERR_OK;
}


// almost an exact duplicate of CompressRGB; just some threes changed to fours
DWORD CodecInst::CompressRGBA(ICCOMPRESS* icinfo, DWORD dwSize) {
  unsigned char* in = (unsigned char*)icinfo->lpInput;
  DWORD* out = (DWORD*)icinfo->lpOutput;
  *out++ = *(DWORD*)in << 8;

  DWORD bits = 0;
  int bit_count_minus_32 = -32;
  unsigned char* end = in + icinfo->lpbiInput->biSizeImage;
  in += 4;
  do {
    HUFF(luma,0,4);
    ++in;
  } while (in < end);

  if (-bit_count_minus_32 < 32)
    *out++ = bits << (-bit_count_minus_32);

  icinfo->lpbiOutput->biSizeImage = DWORD(out) - DWORD(icinfo->lpOutput);
  if (icinfo->lpckid)
    *icinfo->lpckid = FOURCC_HFYU;
  *icinfo->lpdwFlags = AVIIF_KEYFRAME;

  return ICERR_OK;
}


void CodecInst::ConvertRGB24toYUY2(unsigned char* src, unsigned char* dst, int width, int height) {
  const int cyb = int(0.114*219/255*65536+0.5);
  const int cyg = int(0.587*219/255*65536+0.5);
  const int cyr = int(0.299*219/255*65536+0.5);

  for (int row = 0; row < height; ++row) {
    unsigned char* rgb = src + width * 3 * (height-1-row);
    unsigned char* yuv = dst + width * 2 * row;
    for (int col = 0; col < width; col += 2) {
      // y1 and y2 can't overflow
      int y1 = (cyb*rgb[0] + cyg*rgb[1] + cyr*rgb[2] + 0x108000) >> 16;
      yuv[0] = y1;
      int y2 = (cyb*rgb[3] + cyg*rgb[4] + cyr*rgb[5] + 0x108000) >> 16;
      yuv[2] = y2;
      int scaled_y = (y1+y2 - 32) * int(255.0/219.0*32768+0.5);
      int b_y = ((rgb[0]+rgb[3]) << 15) - scaled_y;
      yuv[1] = Clip((b_y >> 10) * int(1/2.018*1024+0.5) + 0x800000);  // u
      int r_y = ((rgb[2]+rgb[5]) << 15) - scaled_y;
      yuv[3] = Clip((r_y >> 10) * int(1/1.596*1024+0.5) + 0x800000);  // v
      yuv += 4;
      rgb += 6;
    }
  }
}


DWORD CodecInst::CompressEnd() {
  if (yuy2_buffer) {
    delete[] yuy2_buffer;
    yuy2_buffer = 0;
  }
  return ICERR_OK;
}


/********************************************************************
********************************************************************/

static bool CanDecompress(LPBITMAPINFOHEADER lpbiIn) {
  int intype = GetBitmapType(lpbiIn);
  return (intype < 0 || intype == 1 || intype == 2);
}


static bool CanDecompress(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  if (!lpbiOut)
    return CanDecompress(lpbiIn);

  // must be 1:1 (no stretching)
  if (lpbiOut && (lpbiOut->biWidth != lpbiIn->biWidth || lpbiOut->biHeight != lpbiIn->biHeight))
    return false;

  int outtype = GetBitmapType(lpbiOut);
  switch (GetBitmapType(lpbiIn)) {
    case -1:
      // YUY2, RGB-24, RGB-32 output for compressed YUY2
      return (outtype == 1 || outtype == 3 || outtype == 4);
    case -2: case 1: case 2:
      // RGB-24, RGB-32 output only for YUY2/UYVY and compressed RGB
      return (outtype == 3 || outtype == 4);
    case -3:
      // RGB-32 output for compressed RGBA
      return (outtype == 4);
    default:
      return false;
  }
}

/********************************************************************
********************************************************************/


DWORD CodecInst::DecompressQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  Msg("DecompressQuery: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));
  return CanDecompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}


// This function should return "the output format which preserves the most
// information."  However, I now provide the option to return RGB format
// instead, since some programs treat the default format as the ONLY format.

DWORD CodecInst::DecompressGetFormat(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {

  // if lpbiOut == NULL, then return the size required to hold an output format struct
  if (lpbiOut == NULL)
    return lpbiIn->biSize;

  if (!CanDecompress(lpbiIn))
    return ICERR_BADFORMAT;

  // get default rgboutput value and allow the user to override
  bool rgboutput = AppFlags() & 1;
  rgboutput = !!GetPrivateProfileInt("debug", "rgboutput", rgboutput, "huffyuv.ini");

  memcpy(lpbiOut, lpbiIn, lpbiIn->biSize);

  int intype = GetBitmapType(lpbiIn);
  if (intype == -3) {
    lpbiOut->biBitCount = 32;   // "RGBA" (though not really)
    lpbiOut->biCompression = 0;
    lpbiOut->biSizeImage = lpbiIn->biWidth * lpbiIn->biHeight * 4;
  } else if (rgboutput || intype == -2 || intype == 1 || intype == 2) {
    lpbiOut->biBitCount = 24;   // RGB
    lpbiOut->biCompression = 0;
    lpbiOut->biSizeImage = lpbiIn->biWidth * lpbiIn->biHeight * 3;
  } else {
    lpbiOut->biBitCount = 16;       // YUY2
    lpbiOut->biCompression = FOURCC_YUY2;
    lpbiOut->biSizeImage = lpbiIn->biWidth * lpbiIn->biHeight * 2;
  }

  return ICERR_OK;
}

DWORD CodecInst::DecompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  InitClip();
  Msg("DecompressBegin: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));
  return CanDecompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}


// decode a single luma/chroma byte from the Huffman-coded bitstream
static inline unsigned char UnhuffLuma(const DWORD* in, unsigned& pos) {
  unsigned word = pos >> 5;
  unsigned bit = pos & 31;
  unsigned val = in[word];
  if (bit)    // the conditional is necessary because shifting by 32 doesn't work correctly on x86 processors
    val = (val << bit) | (in[word+1] >> (32-bit));
  // figure out the appropriate lookup table based on the number of leading zeros
  if (val >= (35 << 23)) {
    pos += decode_luma9[val >> 23].shift;
    return decode_luma9[val >> 23].add;
  }
  if (val >= (15 << 20)) {
    pos += decode_luma12[val >> 20].shift;
    return decode_luma12[val >> 20].add;
  }
  pos += decode_luma17[val >> 15].shift;
  return decode_luma17[val >> 15].add;
}

static inline unsigned char UnhuffChroma(const DWORD* in, unsigned& pos) {
  unsigned word = pos >> 5;
  unsigned bit = pos & 31;
  unsigned val = in[word];
  if (bit)
    val = (val << bit) | (in[word+1] >> (32-bit));
  if (val >= (2 << 23)) {
    pos += decode_chroma9[val >> 23].shift;
    return decode_chroma9[val >> 23].add;
  }
  if (val >= (4 << 16)) {
    pos += decode_chroma16[val >> 16].shift;
    return decode_chroma16[val >> 16].add;
  }
  if (val >= (26 << 9)) {
    pos += decode_chroma23[val >> 9].shift;
    return decode_chroma23[val >> 9].add;
  }
  pos += decode_chroma26[val >> 6].shift;
  return decode_chroma26[val >> 6].add;
}


static inline void YUV2RGB(int y, int u, int v, BYTE* out) {
  const int crv = int(1.596*65536+0.5);
  const int cgv = int(0.813*65536+0.5);
  const int cgu = int(0.391*65536+0.5);
  const int cbu = int(2.018*65536+0.5);

  int scaled_y = (y - 16) * int((255.0/219.0)*65536+0.5);

  out[0] = Clip(scaled_y + (u-128) * cbu); // blue
  out[1] = Clip(scaled_y - (u-128) * cgu - (v-128) * cgv); // green
  out[2] = Clip(scaled_y + (v-128) * crv); // red
}


DWORD CodecInst::Decompress(ICDECOMPRESS* icinfo, DWORD dwSize) {
  int intype = GetBitmapType(icinfo->lpbiInput);
  int outtype = GetBitmapType(icinfo->lpbiOutput);
  if (intype == -1 && (outtype == 1 || outtype >= 3))
    return DecompressYUY2(icinfo, dwSize, outtype > 1 ? outtype : 0);
  else if (intype == -2 && outtype >= 3)
    return DecompressRGB(icinfo, dwSize, outtype);
  else if (intype == -3 && outtype == 4)
    return DecompressRGB(icinfo, dwSize, outtype);
  else if (intype == 1 && outtype >= 3)
    return ConvertYUVtoRGB(icinfo, dwSize, outtype, false);
  else if (intype == 2 && outtype >= 3)
    return ConvertYUVtoRGB(icinfo, dwSize, outtype, true);
  else
    return ICERR_BADFORMAT;
}


DWORD CodecInst::ConvertYUVtoRGB(ICDECOMPRESS* icinfo, DWORD dwSize, int rgb_bytes, bool uyvy) {

  int width = icinfo->lpbiOutput->biWidth;
  int height = icinfo->lpbiOutput->biHeight;
  int size = width * height * rgb_bytes;
  icinfo->lpbiOutput->biSizeImage = size;

  unsigned char* const yuv_base = (unsigned char*)icinfo->lpInput;
  unsigned char* const rgb_base = (unsigned char*)icinfo->lpOutput;
  for (int row = 0; row < height; ++row) {
    unsigned char* y = yuv_base + width * 2 * (height-1-row) + uyvy;
    unsigned char* uv = (unsigned char*)(int(y)^1);
    unsigned char* rgb = rgb_base + width * rgb_bytes * row;
    YUV2RGB(y[0], uv[0], uv[2], rgb);
    YUV2RGB(y[2], uv[0], uv[2], rgb+rgb_bytes);
    for (int col = 2; col < width; col += 2) {
      y += 4;
      uv += 4;
      rgb += rgb_bytes*2;
      // interpolate chroma values
      YUV2RGB(y[0], (int(uv[0])+int(uv[-4]))/2, (int(uv[2])+int(uv[-2]))/2, rgb);
      YUV2RGB(y[2], uv[0], uv[2], rgb+rgb_bytes);
    }
  }
  return ICERR_OK;
}


DWORD CodecInst::DecompressYUY2(ICDECOMPRESS* icinfo, DWORD dwSize, int rgb_bytes) {

  int width = icinfo->lpbiOutput->biWidth;
  int height = icinfo->lpbiOutput->biHeight;
  int size = width * height * (rgb_bytes ? rgb_bytes : 2);
  icinfo->lpbiOutput->biSizeImage = size;

  const DWORD* const in = (DWORD*)icinfo->lpInput;
  unsigned char* const end = (unsigned char*)icinfo->lpOutput + size;
  unsigned char* out = end - width*height*2;

  *(DWORD*)out = in[0];
  out += 4;
  unsigned pos = 32;
  while (out < end) {
    out[0] = out[-2] + UnhuffLuma(in, pos);
    out[1] = out[-3] + UnhuffChroma(in, pos);
    out += 2;
  }

  // convert YUYV to RGB in-place.  This would be easy, were it not for the
  // fact that YUYV is top-to-bottom and RGB is bottom-to-top.  As it is,
  // it's tricky.  If you see a way to avoid the memcpy let me know.
  if (rgb_bytes) {
    unsigned char* const rgb_base = (unsigned char*)icinfo->lpOutput;
    unsigned char* const yuv_base = end - width*height*2;
    for (int row = 0; row < height; ++row) {
      bool first_half = (row < height/2);
      unsigned char* yuv = yuv_base + width * 2 * (first_half ? height-1-row : row);
      unsigned char* rgb = rgb_base + width * rgb_bytes * row;
      YUV2RGB(yuv[0], yuv[1], yuv[3], rgb);
      YUV2RGB(yuv[2], yuv[1], yuv[3], rgb+rgb_bytes);
      for (int col = 2; col < width; col += 2) {
        yuv += 4;
        rgb += rgb_bytes*2;
        // interpolate chroma values
        YUV2RGB(yuv[0], (int(yuv[1])+int(yuv[-3]))/2, (int(yuv[3])+int(yuv[-1]))/2, rgb);
        YUV2RGB(yuv[2], yuv[1], yuv[3], rgb+rgb_bytes);
      }
      if (first_half)
        memcpy(yuv_base + width*2*(height-1-row), yuv_base + width*2*row, width*2);
    }
  }

  return ICERR_OK;
}


DWORD CodecInst::DecompressRGB(ICDECOMPRESS* icinfo, DWORD dwSize, int rgb_bytes) {

  int width = icinfo->lpbiOutput->biWidth;
  int height = icinfo->lpbiOutput->biHeight;
  int size = width * height * rgb_bytes;
  icinfo->lpbiOutput->biSizeImage = size;

  const DWORD* const in = (DWORD*)icinfo->lpInput;
  unsigned char* const end = (unsigned char*)icinfo->lpOutput + size;
  unsigned char* out = (unsigned char*)icinfo->lpOutput;

  *(DWORD*)out = in[0] >> 8;
  unsigned pos = 32;
  if (icinfo->lpbiInput->biBitCount == 32) {
    // RGBA
    out += 4;
    while (out < end) {
      out[0] = out[-4] + UnhuffLuma(in, pos);
      ++out;
    }
  } else if (icinfo->lpbiOutput->biBitCount == 24) {
    // RGB 24-bit
    out += 3;
    while (out < end) {
      out[0] = out[-3] + UnhuffLuma(in, pos);
      ++out;
    }
  } else {
    // RGB 32-bit (pad with zeroes)
    out += 4;
    while (out < end) {
      out[0] = out[-4] + UnhuffLuma(in, pos);
      out[1] = out[-3] + UnhuffLuma(in, pos);
      out[2] = out[-2] + UnhuffLuma(in, pos);
      out[3] = 0;
      out += 4;
    }
  }
  return ICERR_OK;
}


// palette-mapped output only
DWORD CodecInst::DecompressGetPalette(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  return ICERR_BADFORMAT;
}


DWORD CodecInst::DecompressEnd() {
    return ICERR_OK;
}

